﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using NHibernate.Engine;
using NHibernate.Metadata;
using NHibernate.Param;
using NHibernate.Type;

namespace NHibernate.Linq.Visitors
{
	/// <summary>
	/// Locates constants in the expression tree and generates parameters for each one
	/// </summary>
    public class ExpressionParameterVisitor : ContextualExpressionTreeVisitor
	{
		private readonly Dictionary<ConstantExpression, NamedParameter> _parameters = new Dictionary<ConstantExpression, NamedParameter>();
		private readonly ISessionFactory _sessionFactory;

		public ExpressionParameterVisitor(ISessionFactory sessionFactory)
		{
			_sessionFactory = sessionFactory;
		}

		public static IDictionary<ConstantExpression, NamedParameter> Visit(Expression expression, ISessionFactory sessionFactory)
		{
			var visitor = new ExpressionParameterVisitor(sessionFactory);
			
			visitor.VisitExpression(expression);

			return visitor._parameters;
		}

		protected override Expression VisitMethodCallExpression(MethodCallExpression expression)
		{
			if (VisitorUtil.IsDynamicComponentDictionaryGetter(expression, _sessionFactory))
			{
				return expression;
			}
			return base.VisitMethodCallExpression(expression);
		}

		protected override Expression VisitConstantExpression(ConstantExpression expression)
		{
			if (!_parameters.ContainsKey(expression) && !typeof(IQueryable).IsAssignableFrom(expression.Type) && !IsNullObject(expression))
			{
                Tuple<object, IType> valueAndType = GuessValueAndTypeFromContext(expression);
                _parameters.Add(expression, new NamedParameter("p" + (_parameters.Count + 1), valueAndType.First, valueAndType.Second));
			}

			return base.VisitConstantExpression(expression);
		}


        private Tuple<object, IType> GuessValueAndTypeFromContext(ConstantExpression expression)
        {
            // We're trying to determine the type of this parameter.  If we're comparing with a property with a known type,
            // we can check the property's type against what we have.  If they match, we can use it.  Otherwise, we just
            // guess the type.
            object value = expression.Value;

            if (Context.Parent.Expression is BinaryExpression)
            {
                BinaryExpression binaryExpression = Context.Parent.Expression as BinaryExpression;

                if (binaryExpression.NodeType == ExpressionType.Equal ||
                    binaryExpression.NodeType == ExpressionType.NotEqual ||
                    binaryExpression.NodeType == ExpressionType.GreaterThan ||
                    binaryExpression.NodeType == ExpressionType.GreaterThanOrEqual ||
                    binaryExpression.NodeType == ExpressionType.LessThan ||
                    binaryExpression.NodeType == ExpressionType.LessThanOrEqual)
                {
                    Expression otherSide = binaryExpression.Left == expression ? binaryExpression.Right : binaryExpression.Left;

                    if (otherSide.NodeType == ExpressionType.Convert)
                    {
                        Expression preConvertExpression = ((UnaryExpression)otherSide).Operand;
                        otherSide = preConvertExpression;
                    }

                    if (otherSide is MemberExpression)
                    {
                        MemberExpression memberExpression = otherSide as MemberExpression;
                        IClassMetadata metadata = _sessionFactory.GetClassMetadata(memberExpression.Expression.Type);
                        if (metadata == null)
                        {
                            string[] implementors = ((ISessionFactoryImplementor)_sessionFactory).GetImplementors(memberExpression.Expression.Type.FullName);
                            if (implementors.Length == 1)
                                metadata = _sessionFactory.GetClassMetadata(implementors[0]);
                        }
                        if (metadata != null)
                        {
                            IType potentialType = null;
                            string memberName = memberExpression.Member.Name;
                            if (metadata.HasIdentifierProperty && metadata.IdentifierPropertyName == memberName)
                            {
                                potentialType = metadata.IdentifierType;
                            }
                            else
                            {
                                int propertyIndex = Array.IndexOf(metadata.PropertyNames, memberName);
                                if (propertyIndex != -1)
                                {
                                    potentialType = metadata.PropertyTypes[propertyIndex];
                                }
                            }

                            if (potentialType != null)
                            {
                                // This check is possibly too restrictive.
                                if (expression.Type.IsAssignableFrom(potentialType.ReturnedClass))
                                {
                                    if (value is Enum)
                                        value = Enum.ToObject(value.GetType(), value);

                                    return new Tuple<object, IType> { First = value, Second = potentialType };
                                }
                            }
                        }
                    }
                }
            }

            return new Tuple<object, IType>
            {
                First = value,
                Second = value == null ? NHibernateUtil.GuessType(expression.Type) : null
            };
        }
        
        private static bool IsNullObject(ConstantExpression expression)
		{
			return expression.Type == typeof(Object) && expression.Value == null;
		}
	}
}